/* eslint-disable */
module.exports = {
  // Windows
  bs_ie_9: {
    'browser' : 'ie',
    'browser_version' : '9.0',
    'os' : 'Windows',
    'os_version' : '7'
  },
  bs_ie_10: {
    'browser' : 'ie',
    'browser_version' : '10.0',
    'os' : 'Windows',
    'os_version' : '7'
  },
  bs_ie_11: {
    'browser' : 'ie',
    'browser_version' : '11.0',
    'os' : 'Windows',
    'os_version' : '7'
  },
  bs_edge: {
    'browser' : 'edge',
    'browser_version' : '12.0',
    'os' : 'Windows',
    'os_version' : '10'
  },
  // Mac
  bs_safari_6: {
   'browser' : 'safari',
   'browser_version' : '6.2',
   'os' : 'OS X',
   'os_version' : 'Mountain Lion'
  },
  bs_safari_7: {
   'browser' : 'safari',
   'browser_version' : '7.1',
   'os' : 'OS X',
   'os_version' : 'Mavericks'
  },
  bs_safari_8: {
   'browser' : 'safari',
   'browser_version' : '8.0',
   'os' : 'OS X',
   'os_version' : 'Yosemite'
  },
  bs_firefox: {
   'browser' : 'firefox',
   'os' : 'OS X',
   'os_version' : 'Yosemite'
  },
  bs_iphone_5: {
    'browserName' : 'iPhone',
    'device' : 'iPhone 5',
    'os_version': '6.0',
    'os': 'ios'
  },
  bs_iphone_6: {
    'browserName' : 'iPhone',
    'device' : 'iPhone 6',
    'os_version': '8.0',
    'os': 'ios'
  },
  // Android
  bs_android_4: {
    'browserName' : 'android',
    'device' : 'Samsung Galaxy S5',
    'os_version': '4.4',
    'os': 'android'
  }
}